# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import unittest

import lsst.utils
import lsst.utils.tests
from lsst.utils.ellipsis import Ellipsis, EllipsisType


class EllipsisTestCase(lsst.utils.tests.TestCase):
    """Test ellipsis handling."""

    def test_ellipsis(self):
        # These are true at runtime because of typing.TYPE_CHECKING guards in
        # the module.  When MyPy or other type-checkers run, these assertions
        # would not be true, and `Ellipsis` must be used instead of the literal
        # `...` to be understood by mypy.
        self.assertIs(Ellipsis, ...)
        self.assertIs(EllipsisType, type(...))


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.
#

import itertools
import unittest

from lsst.utils.iteration import chunk_iterable, ensure_iterable, isplit


class IterationTestCase(unittest.TestCase):
    """Tests for `iterable` helper."""

    def testIterable(self):
        test_data = (
            # Boolean indicates to test that we know it's
            # meant to be returned unchanged.
            (0, False),
            ("hello", False),
            ({1: 2, 3: 4}, False),
            ([0, 1, 2], True),
            (["hello", "world"], True),
            ({"a", "b", "c"}, True),
        )

        for data, is_iterable in test_data:
            iterator = ensure_iterable(data)
            if not is_iterable:
                # Turn into a list so we can check against the
                # expected result.
                data = [data]
            for original, from_iterable in itertools.zip_longest(data, iterator):
                self.assertEqual(original, from_iterable)

    def testChunking(self):
        """Chunk iterables."""
        simple = list(range(101))
        out = []
        n_chunks = 0
        for chunk in chunk_iterable(simple, chunk_size=10):
            out.extend(chunk)
            n_chunks += 1
        self.assertEqual(out, simple)
        self.assertEqual(n_chunks, 11)

        test_dict = {k: 1 for k in range(101)}
        n_chunks = 0
        for chunk in chunk_iterable(test_dict, chunk_size=45):
            # Subtract 1 for each key in chunk
            for k in chunk:
                test_dict[k] -= 1
            n_chunks += 1
        # Should have matched every key
        self.assertEqual(sum(test_dict.values()), 0)
        self.assertEqual(n_chunks, 3)

    def testIsplit(self):
        # Test compatibility with str.split
        seps = ("\n", " ", "d")
        input_str = "ab\ncd ef\n"

        for sep in seps:
            for input in (input_str, input_str.encode()):
                test_sep = sep.encode() if isinstance(input, bytes) else sep
                isp = list(isplit(input, sep=test_sep))
                ssp = input.split(test_sep)
                self.assertEqual(isp, ssp)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

import sys
import types

import numpy as np

__all__ = ("continueClass", "inClass", "TemplateMeta")


INTRINSIC_SPECIAL_ATTRIBUTES = frozenset(
    (
        "__qualname__",
        "__module__",
        "__metaclass__",
        "__dict__",
        "__weakref__",
        "__class__",
        "__subclasshook__",
        "__name__",
        "__doc__",
    )
)


def isAttributeSafeToTransfer(name, value):
    """Return True if an attribute is safe to monkeypatch-transfer to another
    class.

    This rejects special methods that are defined automatically for all
    classes, leaving only those explicitly defined in a class decorated by
    `continueClass` or registered with an instance of `TemplateMeta`.
    """
    if name.startswith("__") and (
        value is getattr(object, name, None) or name in INTRINSIC_SPECIAL_ATTRIBUTES
    ):
        return False
    return True


def continueClass(cls):
    """Re-open the decorated class, adding any new definitions into the
    original.

    For example:

    .. code-block:: python

        class Foo:
            pass

        @continueClass
        class Foo:
            def run(self):
                return None

    is equivalent to:

    .. code-block:: python

        class Foo:
            def run(self):
                return None

    .. warning::

        Python's built-in `super` function does not behave properly in classes
        decorated with `continueClass`.  Base class methods must be invoked
        directly using their explicit types instead.

    """
    orig = getattr(sys.modules[cls.__module__], cls.__name__)
    for name in dir(cls):
        # Common descriptors like classmethod and staticmethod can only be
        # accessed without invoking their magic if we use __dict__; if we use
        # getattr on those we'll get e.g. a bound method instance on the dummy
        # class rather than a classmethod instance we can put on the target
        # class.
        attr = cls.__dict__.get(name, None) or getattr(cls, name)
        if isAttributeSafeToTransfer(name, attr):
            setattr(orig, name, attr)
    return orig


def inClass(cls, name=None):
    """Add the decorated function to the given class as a method.

    For example:

    .. code-block:: python

        class Foo:
            pass

        @inClass(Foo)
        def run(self):
            return None

    is equivalent to:

    .. code-block:: python

        class Foo:
            def run(self):
                return None

    Standard decorators like ``classmethod``, ``staticmethod``, and
    ``property`` may be used *after* this decorator.  Custom decorators
    may only be used if they return an object with a ``__name__`` attribute
    or the ``name`` optional argument is provided.
    """

    def decorate(func):
        # Using 'name' instead of 'name1' breaks the closure because
        # assignment signals a strictly local variable.
        name1 = name
        if name1 is None:
            if hasattr(func, "__name__"):
                name1 = func.__name__
            else:
                if hasattr(func, "__func__"):
                    # classmethod and staticmethod have __func__ but
                    # no __name__
                    name1 = func.__func__.__name__
                elif hasattr(func, "fget"):
                    # property has fget but no __name__
                    name1 = func.fget.__name__
                else:
                    raise ValueError(f"Could not guess attribute name for '{func}'.")
        setattr(cls, name1, func)
        return func

    return decorate


class TemplateMeta(type):
    """A metaclass for abstract base classes that tie together wrapped C++
    template types.

    C++ template classes are most easily wrapped with a separate Python class
    for each template type, which results in an unnatural Python interface.
    TemplateMeta provides a thin layer that connects these Python classes by
    giving them a common base class and acting as a factory to construct them
    in a consistent way.

    To use, simply create a new class with the name of the template class, and
    use ``TemplateMeta`` as its metaclass, and then call ``register`` on each
    of its subclasses.  This registers the class with a "type key" - usually a
    Python representation of the C++ template types.  The type key must be a
    hashable object - strings, type objects, and tuples of these (for C++
    classes with multiple template parameters) are good choices.  Alternate
    type keys for existing classes can be added by calling ``alias``, but only
    after a subclass already been registered with a "primary" type key.  For
    example::

    .. code-block:: python

        import numpy as np
        from ._image import ImageF, ImageD

        class Image(metaclass=TemplateMeta):
            pass

        Image.register(np.float32, ImageF)
        Image.register(np.float64, ImageD)
        Image.alias("F", ImageF)
        Image.alias("D", ImageD)

    We have intentionally used ``numpy`` types as the primary keys for these
    objects in this example, with strings as secondary aliases simply because
    the primary key is added as a ``dtype`` attribute on the the registered
    classes (so ``ImageF.dtype == numpy.float32`` in the above example).

    This allows user code to construct objects directly using ``Image``, as
    long as an extra ``dtype`` keyword argument is passed that matches one of
    the type keys::

    .. code-block:: python

        img = Image(52, 64, dtype=np.float32)

    This simply forwards additional positional and keyword arguments to the
    wrapped template class's constructor.

    The choice of "dtype" as the name of the template parameter is also
    configurable, and in fact multiple template parameters are also supported,
    by setting a ``TEMPLATE_PARAMS`` class attribute on the ABC to a tuple
    containing the names of the template parameters.  A ``TEMPLATE_DEFAULTS``
    attribute can also be defined to a tuple of the same length containing
    default values for the template parameters, allowing them to be omitted in
    constructor calls.  When the length of these attributes is more than one,
    the type keys passed to ``register`` and ``alias`` should be tuple of the
    same length; when the length of these attributes is one, type keys should
    generally not be tuples.

    As an aid for those writing the Python wrappers for C++ classes,
    ``TemplateMeta`` also provides a way to add pure-Python methods and other
    attributes to the wrapped template classes.  To add a ``sum`` method to
    all registered types, for example, we can just do::

    .. code-block:: python

        class Image(metaclass=TemplateMeta):

            def sum(self):
                return np.sum(self.getArray())

        Image.register(np.float32, ImageF)
        Image.register(np.float64, ImageD)

    .. note::

        ``TemplateMeta`` works by overriding the ``__instancecheck__`` and
        ``__subclasscheck__`` special methods, and hence does not appear in
        its registered subclasses' method resolution order or ``__bases__``
        attributes. That means its attributes are not inherited by registered
        subclasses. Instead, attributes added to an instance of
        ``TemplateMeta`` are *copied* into the types registered with it. These
        attributes will thus *replace* existing attributes in those classes
        with the same name, and subclasses cannot delegate to base class
        implementations of these methods.

    Finally, abstract base classes that use ``TemplateMeta`` define a dict-
    like interface for accessing their registered subclasses, providing
    something like the C++ syntax for templates::

    .. code-block:: python

        Image[np.float32] -> ImageF
        Image["D"] -> ImageD

    Both primary dtypes and aliases can be used as keys in this interface,
    which means types with aliases will be present multiple times in the dict.
    To obtain the sequence of unique subclasses, use the ``__subclasses__``
    method.

    .. warning::

        Python's built-in `super` function does not behave properly in classes
        that have `TemplateMeta` as their metaclass (which should be rare, as
        TemplateMeta ABCs will have base classes of their own)..

    """

    def __new__(cls, name, bases, attrs):
        # __new__ is invoked when the abstract base class is defined (via a
        # class statement).  We save a dict of class attributes (including
        # methods) that were defined in the class body so we can copy them
        # to registered subclasses later.
        # We also initialize an empty dict to store the registered subclasses.
        attrs["_inherited"] = {k: v for k, v in attrs.items() if isAttributeSafeToTransfer(k, v)}
        # The special "TEMPLATE_PARAMS" class attribute, if defined, contains
        # names of the template parameters, which we use to set those
        # attributes on registered subclasses and intercept arguments to the
        # constructor.  This line removes it from the dict of things that
        # should be inherited while setting a default of 'dtype' if it's not
        # defined.
        attrs["TEMPLATE_PARAMS"] = attrs["_inherited"].pop("TEMPLATE_PARAMS", ("dtype",))
        attrs["TEMPLATE_DEFAULTS"] = attrs["_inherited"].pop(
            "TEMPLATE_DEFAULTS", (None,) * len(attrs["TEMPLATE_PARAMS"])
        )
        attrs["_registry"] = {}
        self = type.__new__(cls, name, bases, attrs)

        if len(self.TEMPLATE_PARAMS) == 0:
            raise ValueError("TEMPLATE_PARAMS must be a tuple with at least one element.")
        if len(self.TEMPLATE_DEFAULTS) != len(self.TEMPLATE_PARAMS):
            raise ValueError("TEMPLATE_PARAMS and TEMPLATE_DEFAULTS must have same length.")
        return self

    def __call__(cls, *args, **kwds):
        # __call__ is invoked when someone tries to construct an instance of
        # the abstract base class.
        # If the ABC defines a "TEMPLATE_PARAMS" attribute, we use those
        # strings as the kwargs we should intercept to find the right type.
        # Generate a type mapping key from input keywords. If the type returned
        # from the keyword lookup is a numpy dtype object, fetch the underlying
        # type of the dtype
        key = []
        for p, d in zip(cls.TEMPLATE_PARAMS, cls.TEMPLATE_DEFAULTS):
            tempKey = kwds.pop(p, d)
            if isinstance(tempKey, np.dtype):
                tempKey = tempKey.type
            key.append(tempKey)
        key = tuple(key)

        # indices are only tuples if there are multiple elements
        clz = cls._registry.get(key[0] if len(key) == 1 else key, None)
        if clz is None:
            d = dict(zip(cls.TEMPLATE_PARAMS, key))
            raise TypeError(f"No registered subclass for {d}.")
        return clz(*args, **kwds)

    def __subclasscheck__(cls, subclass):
        # Special method hook for the issubclass built-in: we return true for
        # any registered type or true subclass thereof.
        if subclass in cls._registry.values():
            return True
        return any(issubclass(subclass, v) for v in cls._registry.values())

    def __instancecheck__(cls, instance):
        # Special method hook for the isinstance built-in: we return true for
        # an instance of any registered type or true subclass thereof.
        if type(instance) in cls._registry.values():
            return True
        return any(isinstance(instance, v) for v in cls._registry.values())

    def __subclasses__(cls):
        """Return a tuple of all classes that inherit from this class."""
        # This special method isn't defined as part of the Python data model,
        # but it exists on builtins (including ABCMeta), and it provides useful
        # functionality.
        return tuple(set(cls._registry.values()))

    def register(cls, key, subclass):
        """Register a subclass of this ABC with the given key (a string,
        number, type, or other hashable).

        Register may only be called once for a given key or a given subclass.
        """
        if key is None:
            raise ValueError("None may not be used as a key.")
        if subclass in cls._registry.values():
            raise ValueError("This subclass has already registered with another key; use alias() instead.")
        if cls._registry.setdefault(key, subclass) != subclass:
            if len(cls.TEMPLATE_PARAMS) == 1:
                d = {cls.TEMPLATE_PARAMS[0]: key}
            else:
                d = dict(zip(cls.TEMPLATE_PARAMS, key))
            raise KeyError(f"Another subclass is already registered with {d}")
        # If the key used to register a class matches the default key,
        # make the static methods available through the ABC
        if cls.TEMPLATE_DEFAULTS:
            defaults = cls.TEMPLATE_DEFAULTS[0] if len(cls.TEMPLATE_DEFAULTS) == 1 else cls.TEMPLATE_DEFAULTS
            if key == defaults:
                conflictStr = (
                    "Base class has attribute {}"
                    " which is a {} method of {}."
                    " Cannot link method to base class."
                )
                # In the following if statements, the explicit lookup in
                # __dict__ must be done, as a call to getattr returns the
                # bound method, which no longer reports as a static or class
                # method. The static methods must be transfered to the ABC
                # in this unbound state, so that python will still see them
                # as static methods and not attempt to pass self. The class
                # methods must be transfered to the ABC as a bound method
                # so that the correct cls be called with the class method
                for name in subclass.__dict__:
                    if name in ("__new__", "__init_subclass__"):
                        continue
                    obj = subclass.__dict__[name]
                    # copy over the static methods
                    isBuiltin = isinstance(obj, types.BuiltinFunctionType)
                    isStatic = isinstance(obj, staticmethod)
                    if isBuiltin or isStatic:
                        if hasattr(cls, name):
                            raise AttributeError(conflictStr.format(name, "static", subclass))
                        setattr(cls, name, obj)
                    # copy over the class methods
                    elif isinstance(obj, classmethod):
                        if hasattr(cls, name):
                            raise AttributeError(conflictStr.format(name, "class", subclass))
                        setattr(cls, name, getattr(subclass, name))

        def setattrSafe(name, value):
            try:
                currentValue = getattr(subclass, name)
                if currentValue != value:
                    msg = "subclass already has a '{}' attribute with value {} != {}."
                    raise ValueError(msg.format(name, currentValue, value))
            except AttributeError:
                setattr(subclass, name, value)

        if len(cls.TEMPLATE_PARAMS) == 1:
            setattrSafe(cls.TEMPLATE_PARAMS[0], key)
        elif len(cls.TEMPLATE_PARAMS) == len(key):
            for p, k in zip(cls.TEMPLATE_PARAMS, key):
                setattrSafe(p, k)
        else:
            raise ValueError(
                f"key must have {len(cls.TEMPLATE_PARAMS)} elements (one for each of {cls.TEMPLATE_PARAMS})"
            )

        for name, attr in cls._inherited.items():
            setattr(subclass, name, attr)

    def alias(cls, key, subclass):
        """Add an alias that allows an existing subclass to be accessed with a
        different key.
        """
        if key is None:
            raise ValueError("None may not be used as a key.")
        if key in cls._registry:
            raise KeyError(f"Cannot multiply-register key {key}")
        primaryKey = tuple(getattr(subclass, p, None) for p in cls.TEMPLATE_PARAMS)
        if len(primaryKey) == 1:
            # indices are only tuples if there are multiple elements
            primaryKey = primaryKey[0]
        if cls._registry.get(primaryKey, None) != subclass:
            raise ValueError("Subclass is not registered with this base class.")
        cls._registry[key] = subclass

    # Immutable mapping interface defined below.  We don't use collections
    # mixins because we don't want their comparison operators.

    def __getitem__(cls, key):
        return cls._registry[key]

    def __iter__(cls):
        return iter(cls._registry)

    def __len__(cls):
        return len(cls._registry)

    def __contains__(cls, key):
        return key in cls._registry

    def keys(cls):
        """Return an iterable containing all keys (including aliases)."""
        return cls._registry.keys()

    def values(cls):
        """Return an iterable of registered subclasses, with duplicates
        corresponding to any aliases.
        """
        return cls._registry.values()

    def items(cls):
        """Return an iterable of (key, subclass) pairs."""
        return cls._registry.items()

    def get(cls, key, default=None):
        """Return the subclass associated with the given key (including
        aliases), or ``default`` if the key is not recognized.
        """
        return cls._registry.get(key, default)



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("inheritDoc",)

from collections.abc import Callable


def inheritDoc(klass: type) -> Callable:
    """Extend existing documentation for a method that exists in another
    class and extend it with any additional documentation defined.

    This decorator takes a class from which to draw documentation from as an
    argument. This is so that any class may be used as a source of
    documentation and not just the immediate parent of a class. This is useful
    when there may be a long inheritance chain, or in the case of mixins.

    Parameters
    ----------
    klass : object
        The class to inherit documentation from.

    Returns
    -------
    decorator : callable
        Intermediate decorator used in the documentation process.
    """

    def tmpDecorator(method: type) -> Callable:
        """Update the documentation from a class with the same method."""
        methodName = method.__name__
        if not hasattr(klass, methodName):
            raise AttributeError(f"{klass} has no method named {methodName} to inherit from")
        appendText = method.__doc__ or ""
        method.__doc__ = getattr(klass, methodName).__doc__ + appendText
        return method

    return tmpDecorator



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

"""
A type-annotation workaround for ``...`` not having an exposed type in Python
prior to version 3.10.

This module provides ``Ellipsis`` and ``EllipsisType`` symbols that are
conditionally defined to point to the built-in "``...``" singleton and its type
(respectively) at runtime, and an enum class and instance if
`typing.TYPE_CHECKING` is `True` (an approach first suggested
`here <https://github.com/python/typing/issues/684#issuecomment-548203158>`_).
Type checkers should recognize enum literals as singletons, making this pair
behave as expected under ``is`` comparisons and type-narrowing expressions,
such as::

    v: EllipsisType | int
    if v is not Ellipsis:
        v += 2  # type checker should now see v as a pure int

This works best when there is a clear boundary between code that needs to be
type-checked and can use  ``Ellipsis`` instead of a literal "``...``", and
calling code that is either not type-checked or uses `typing.Any` to accept
literal "``...``" values.
"""

from __future__ import annotations

__all__ = ("Ellipsis", "EllipsisType")

import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from enum import Enum

    class EllipsisType(Enum):
        """The type associated with an `...`"""

        Ellipsis = "..."

    Ellipsis = EllipsisType.Ellipsis

else:
    try:
        # Present in Python >= 3.10
        from types import EllipsisType

        # If EllipsisType is defined then produce a deprecation warning.
        warnings.warn(
            f"Module {__name__} is deprecated for Python 3.10 or later, native type `types.EllipsisType` "
            "should be used instead.",
            DeprecationWarning,
            stacklevel=2,
        )
    except ImportError:
        EllipsisType = type(Ellipsis)
    Ellipsis = Ellipsis



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence

import numpy as np


def calculate_safe_plotting_limits(
    data_series: Sequence,
    percentile: float = 99.9,
    constant_extra: float | None = None,
    symmetric_around_zero: bool = False,
) -> tuple[float, float]:
    """Calculate the right limits for plotting for one or more data series.

    Given one or more data series with potential outliers, calculated the
    values to pass for ymin, ymax so that extreme outliers don't ruin the
    plot. If you are plotting several series on a single axis, pass them
    all in and the overall plotting range will be given.

    Parameters
    ----------
    data_series : `iterable` or `iterable` of `iterable`
        One or more data series which will be going on the same axis, and
        therefore want to have their common plotting limits calculated.

    Returns
    -------
    ymin : `float`
        The value to set the ylim minimum to.
    ymax : `float`
        The value to set the ylim maximum to.
    """
    localFunc = make_calculate_safe_plotting_limits(percentile, constant_extra, symmetric_around_zero)
    return localFunc(data_series)


def make_calculate_safe_plotting_limits(
    percentile: float = 99.9,
    constant_extra: float | None = None,
    symmetric_around_zero: bool = False,
) -> Callable[[Sequence], tuple[float, float]]:
    """Make a ``calculate_safe_plotting_limits`` closure to get the common
    limits when not all data series are available initially.

    Parameters
    ----------
    percentile : `float`, optional
        The percentile used to clip the outliers from the data.
    constant_extra : `float`, optional
        The amount that's added on each side of the range so that data does not
        quite touch the axes. If the default ``None`` is left then 5% of the
        data range is added for cosmetics, but if zero is set this will
        overrides this behaviour and zero you will get.
    symmetric_around_zero : `bool`, optional
        Make the limits symmetric around zero?

    Returns
    -------
    calculate_safe_plotting_limits : `callable`
        The calculate_safe_plotting_limits function to pass the data series to.
    """
    memory: list[Sequence] = []

    def calculate_safe_plotting_limits(
        data_series: Sequence,  # a sequence of sequences is still a sequence
    ) -> tuple[float, float]:
        """Calculate the right limits for plotting for one or more data series.

        Given one or more data series with potential outliers, calculated the
        values to pass for ymin, ymax so that extreme outliers don't ruin the
        plot. If you are plotting several series on a single axis, pass them
        all in and the overall plotting range will be given.

        Parameters
        ----------
        data_series : `iterable` or `iterable` of `iterable`
            One or more data series which will be going on the same axis, and
            therefore want to have their common plotting limits calculated.

        Returns
        -------
        ymin : `float`
            The value to set the ylim minimum to.
        ymax : `float`
            The value to set the ylim maximum to.
        """
        nonlocal constant_extra
        nonlocal percentile
        nonlocal symmetric_around_zero

        if not isinstance(data_series, Iterable):
            raise TypeError("data_series must be either an iterable, or an iterable of iterables")

        # now we're sure we have an iterable, if it's just one make it a list
        # of it lsst.utils.ensure_iterable is not suitable here as we already
        # have one, we would need ensure_iterable_of_iterables here

        # np.array are Iterable but not Sequence so isinstance that
        if not isinstance(data_series[0], Iterable):
            # we have a single data series, not multiple, wrap in [] so we can
            # iterate over it as if we were given many
            data_series = [data_series]

        memory.extend(data_series)

        mins = []
        maxs = []

        for dataSeries in memory:
            max_val = np.nanpercentile(dataSeries, percentile)
            min_val = np.nanpercentile(dataSeries, 100.0 - percentile)

            if constant_extra is None:
                data_range = max_val - min_val
                constant_extra = 0.05 * data_range

            max_val += constant_extra
            min_val -= constant_extra

            maxs.append(max_val)
            mins.append(min_val)

        max_val = max(maxs)
        min_val = min(mins)

        if symmetric_around_zero:
            biggest_abs = max(abs(min_val), abs(max_val))
            return -biggest_abs, biggest_abs

        return min_val, max_val

    return calculate_safe_plotting_limits



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = (
    "TRACE",
    "VERBOSE",
    "getLogger",
    "getTraceLogger",
    "LsstLogAdapter",
    "PeriodicLogger",
    "trace_set_at",
)

import logging
import sys
import time
from collections.abc import Generator
from contextlib import contextmanager
from logging import LoggerAdapter
from typing import Any, Union

try:
    import lsst.log.utils as logUtils
except ImportError:
    logUtils = None


# log level for trace (verbose debug).
TRACE = 5
logging.addLevelName(TRACE, "TRACE")

# Verbose logging is midway between INFO and DEBUG.
VERBOSE = (logging.INFO + logging.DEBUG) // 2
logging.addLevelName(VERBOSE, "VERBOSE")


def _calculate_base_stacklevel(default: int, offset: int) -> int:
    """Calculate the default logging stacklevel to use.

    Parameters
    ----------
    default : `int`
        The stacklevel to use in Python 3.11 and newer where the only
        thing to take into account is the number of levels above the core
        Python logging infrastructure.
    offset : `int`
        The offset to apply for older Python implementations that need to
        take into account internal call stacks.

    Returns
    -------
    stacklevel : `int`
        The stack level to pass to internal logging APIs that should result
        in the log messages being reported in caller lines.

    Notes
    -----
    In Python 3.11 the logging infrastructure was fixed such that we no
    longer need to understand that a LoggerAdapter log messages need to
    have a stack level one higher than a Logger would need. ``stacklevel=1``
    now always means "log from the caller's line" without the caller having
    to understand internal implementation details.
    """
    stacklevel = default
    if sys.version_info < (3, 11, 0):
        stacklevel += offset
    return stacklevel


def trace_set_at(name: str, number: int) -> None:
    """Adjust logging level to display messages with the trace number being
    less than or equal to the provided value.

    Parameters
    ----------
    name : `str`
        Name of the logger.
    number : `int`
        The trace number threshold for display.

    Examples
    --------
    .. code-block:: python

       lsst.utils.logging.trace_set_at("lsst.afw", 3)

    This will set loggers ``TRACE0.lsst.afw`` to ``TRACE3.lsst.afw`` to
    ``DEBUG`` and ``TRACE4.lsst.afw`` and ``TRACE5.lsst.afw`` to ``INFO``.

    Notes
    -----
    Loggers ``TRACE0.`` to ``TRACE5.`` are set. All loggers above
    the specified threshold are set to ``INFO`` and those below the threshold
    are set to ``DEBUG``.  The expectation is that ``TRACE`` loggers only
    issue ``DEBUG`` log messages.

    If ``lsst.log`` is installed, this function will also call
    `lsst.log.utils.traceSetAt` to ensure that non-Python loggers are
    also configured correctly.
    """
    for i in range(6):
        level = logging.INFO if i > number else logging.DEBUG
        getTraceLogger(name, i).setLevel(level)

    # if lsst log is available also set the trace loggers there.
    if logUtils is not None:
        logUtils.traceSetAt(name, number)


class _F:
    """Format, supporting `str.format()` syntax.

    Notes
    -----
    This follows the recommendation from
    https://docs.python.org/3/howto/logging-cookbook.html#using-custom-message-objects
    """

    def __init__(self, fmt: str, /, *args: Any, **kwargs: Any):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.fmt.format(*self.args, **self.kwargs)


class LsstLogAdapter(LoggerAdapter):
    """A special logging adapter to provide log features for LSST code.

    Expected to be instantiated initially by a call to `getLogger()`.

    This class provides enhancements over `logging.Logger` that include:

    * Methods for issuing trace and verbose level log messages.
    * Provision of a context manager to temporarily change the log level.
    * Attachment of logging level constants to the class to make it easier
      for a Task writer to access a specific log level without having to
      know the underlying logger class.
    """

    # Store logging constants in the class for convenience. This is not
    # something supported by Python loggers but can simplify some
    # logic if the logger is available.
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING

    # Python supports these but prefers they are not used.
    FATAL = logging.FATAL
    WARN = logging.WARN

    # These are specific to Tasks
    TRACE = TRACE
    VERBOSE = VERBOSE

    # The stack level to use when issuing log messages. For Python 3.11
    # this is generally 2 (this method and the internal infrastructure).
    # For older python we need one higher because of the extra indirection
    # via LoggingAdapter internals.
    _stacklevel = _calculate_base_stacklevel(2, 1)

    @contextmanager
    def temporary_log_level(self, level: int | str) -> Generator:
        """Temporarily set the level of this logger.

        Parameters
        ----------
        level : `int`
            The new temporary log level.
        """
        old = self.level
        self.setLevel(level)
        try:
            yield
        finally:
            self.setLevel(old)

    @property
    def level(self) -> int:
        """Return current level of this logger (``int``)."""
        return self.logger.level

    def getChild(self, name: str) -> LsstLogAdapter:
        """Get the named child logger.

        Parameters
        ----------
        name : `str`
            Name of the child relative to this logger.

        Returns
        -------
        child : `LsstLogAdapter`
            The child logger.
        """
        return getLogger(name=name, logger=self.logger)

    def _process_stacklevel(self, kwargs: dict[str, Any], offset: int = 0) -> int:
        # Return default stacklevel, taking into account kwargs[stacklevel].
        stacklevel = self._stacklevel
        if "stacklevel" in kwargs:
            # External user expects stacklevel=1 to mean "report from their
            # line" but the code here is already trying to achieve that by
            # default. Therefore if an external stacklevel is specified we
            # adjust their stacklevel request by 1.
            stacklevel = stacklevel + kwargs.pop("stacklevel") - 1

        # The offset can be used to indicate that we have to take into account
        # additional internal layers before calling python logging.
        return _calculate_base_stacklevel(stacklevel, offset)

    def fatal(self, msg: str, *args: Any, **kwargs: Any) -> None:
        # Python does not provide this method in LoggerAdapter but does
        # not formally deprecate it in favor of critical() either.
        # Provide it without deprecation message for consistency with Python.
        # Have to adjust stacklevel on Python 3.10 and older to account
        # for call through self.critical.
        stacklevel = self._process_stacklevel(kwargs, offset=1)
        self.critical(msg, *args, **kwargs, stacklevel=stacklevel)

    def verbose(self, fmt: str, *args: Any, **kwargs: Any) -> None:
        """Issue a VERBOSE level log message.

        Arguments are as for `logging.info`.
        ``VERBOSE`` is between ``DEBUG`` and ``INFO``.
        """
        # There is no other way to achieve this other than a special logger
        # method.
        # Stacklevel is passed in so that the correct line is reported
        stacklevel = self._process_stacklevel(kwargs)
        self.log(VERBOSE, fmt, *args, **kwargs, stacklevel=stacklevel)

    def trace(self, fmt: str, *args: Any, **kwargs: Any) -> None:
        """Issue a TRACE level log message.

        Arguments are as for `logging.info`.
        ``TRACE`` is lower than ``DEBUG``.
        """
        # There is no other way to achieve this other than a special logger
        # method.
        stacklevel = self._process_stacklevel(kwargs)
        self.log(TRACE, fmt, *args, **kwargs, stacklevel=stacklevel)

    def setLevel(self, level: int | str) -> None:
        """Set the level for the logger, trapping lsst.log values.

        Parameters
        ----------
        level : `int`
            The level to use. If the level looks too big to be a Python
            logging level it is assumed to be a lsst.log level.
        """
        if isinstance(level, int) and level > logging.CRITICAL:
            self.logger.warning(
                "Attempting to set level to %d -- looks like an lsst.log level so scaling it accordingly.",
                level,
            )
            level //= 1000

        self.logger.setLevel(level)

    @property
    def handlers(self) -> list[logging.Handler]:
        """Log handlers associated with this logger."""
        return self.logger.handlers

    def addHandler(self, handler: logging.Handler) -> None:
        """Add a handler to this logger.

        The handler is forwarded to the underlying logger.
        """
        self.logger.addHandler(handler)

    def removeHandler(self, handler: logging.Handler) -> None:
        """Remove the given handler from the underlying logger."""
        self.logger.removeHandler(handler)


def getLogger(name: str | None = None, logger: logging.Logger | None = None) -> LsstLogAdapter:
    """Get a logger compatible with LSST usage.

    Parameters
    ----------
    name : `str`, optional
        Name of the logger. Root logger if `None`.
    logger : `logging.Logger` or `LsstLogAdapter`
        If given the logger is converted to the relevant logger class.
        If ``name`` is given the logger is assumed to be a child of the
        supplied logger.

    Returns
    -------
    logger : `LsstLogAdapter`
        The relevant logger.

    Notes
    -----
    A `logging.LoggerAdapter` is used since it is easier to provide a more
    uniform interface than when using `logging.setLoggerClass`. An adapter
    can be wrapped around the root logger and the `~logging.setLoggerClass`
    will return the logger first given that name even if the name was
    used before the `~lsst.pipe.base.Task` was created.
    """
    if not logger:
        logger = logging.getLogger(name)
    elif name:
        logger = logger.getChild(name)
    return LsstLogAdapter(logger, {})


LsstLoggers = Union[logging.Logger, LsstLogAdapter]


def getTraceLogger(logger: str | LsstLoggers, trace_level: int) -> LsstLogAdapter:
    """Get a logger with the appropriate TRACE name.

    Parameters
    ----------
    logger : `logging.Logger` or `LsstLogAdapter` or `lsst.log.Log` or `str`
        A logger to be used to derive the new trace logger. Can be a logger
        object that supports the ``name`` property or a string.
    trace_level : `int`
        The trace level to use for the logger.

    Returns
    -------
    trace_logger : `LsstLogAdapter`
        A new trace logger. The name will be derived by pre-pending ``TRACEn.``
        to the name of the supplied logger. If the root logger is given
        the returned logger will be named ``TRACEn``.
    """
    name = getattr(logger, "name", str(logger))
    trace_name = f"TRACE{trace_level}.{name}" if name else f"TRACE{trace_level}"
    return getLogger(trace_name)


class PeriodicLogger:
    """Issue log messages if a time threshold has elapsed.

    This class can be used in long-running sections of code where it would
    be useful to issue a log message periodically to show that the
    algorithm is progressing.

    Parameters
    ----------
    logger : `logging.Logger` or `LsstLogAdapter`
        Logger to use when issuing a message.
    interval : `float`
        The minimum interval between log messages. If `None` the class
        default will be used.
    level : `int`, optional
        Log level to use when issuing messages.
    """

    LOGGING_INTERVAL = 600.0
    """Default interval between log messages."""

    def __init__(self, logger: LsstLoggers, interval: float | None = None, level: int = VERBOSE):
        self.logger = logger
        self.interval = interval if interval is not None else self.LOGGING_INTERVAL
        self.level = level
        self.next_log_time = time.time() + self.interval
        self.num_issued = 0

        # The stacklevel we need to issue logs is determined by the type
        # of logger we have been given. A LoggerAdapter has an extra
        # level of indirection. In Python 3.11 the logging infrastructure
        # takes care to check for internal logging stack frames so there
        # is no need for a difference.
        self._stacklevel = _calculate_base_stacklevel(2, 1 if isinstance(self.logger, LoggerAdapter) else 0)

    def log(self, msg: str, *args: Any) -> bool:
        """Issue a log message if the interval has elapsed.

        Parameters
        ----------
        msg : `str`
            Message to issue if the time has been exceeded.
        *args : Any
            Parameters to be passed to the log system.

        Returns
        -------
        issued : `bool`
            Returns `True` if a log message was sent to the logging system.
            Returns `False` if the interval has not yet elapsed. Returning
            `True` does not indicate whether the log message was in fact
            issued by the logging system.
        """
        if (current_time := time.time()) > self.next_log_time:
            self.logger.log(self.level, msg, *args, stacklevel=self._stacklevel)
            self.next_log_time = current_time + self.interval
            self.num_issued += 1
            return True
        return False



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import unittest

from astropy import units as u
from lsst.utils.usage import get_current_mem_usage, get_peak_mem_usage


class UsageTestCase(unittest.TestCase):
    """Test resource usage functions."""

    def testGetCurrentMemUsage(self):
        main, children = get_current_mem_usage()
        self.assertGreater(main, 0 * u.byte)
        self.assertGreaterEqual(children, 0 * u.byte)

        self.assertTrue(main.unit.is_equivalent(u.byte))
        self.assertTrue(children.unit.is_equivalent(u.byte))

    def testGetPeakMemUsage(self):
        main, child = get_peak_mem_usage()
        self.assertGreater(main, 0 * u.byte)
        self.assertGreaterEqual(child, 0 * u.byte)

        self.assertTrue(main.unit.is_equivalent(u.byte))
        self.assertTrue(child.unit.is_equivalent(u.byte))


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import logging
import time
import unittest

from lsst.utils.logging import PeriodicLogger, getLogger, trace_set_at


class TestLogging(unittest.TestCase):
    """Simple unit tests for Task logging."""

    def testLogLevels(self):
        """Check that the new log levels look reasonable."""
        root = getLogger()

        self.assertEqual(root.DEBUG, logging.DEBUG)
        self.assertGreater(root.VERBOSE, logging.DEBUG)
        self.assertLess(root.VERBOSE, logging.INFO)
        self.assertLess(root.TRACE, logging.DEBUG)

    def testLogCommands(self):
        """Check that all the log commands work."""
        root = getLogger()

        with self.assertLogs(level=root.TRACE) as cm:
            root.trace("Trace")
            root.debug("Debug")
            root.verbose("Verbose")
            root.verbose("Verbose with stacklevel", stacklevel=1)
            root.info("Info")
            root.warning("Warning")
            root.fatal("Fatal")
            root.critical("Critical")
            root.error("Error")

        self.assertEqual(len(cm.records), 9)

        # Check that each record has an explicit level name rather than
        # "Level N" and comes from this file (and not the logging.py).
        for record in cm.records:
            self.assertRegex(record.levelname, "^[A-Z]+$")
            self.assertEqual(record.filename, "test_logging.py")

        with self.assertLogs(level=root.DEBUG) as cm:
            # Should only issue the INFO message.
            with root.temporary_log_level(root.INFO):
                root.info("Info")
                root.debug("Debug")
        self.assertEqual(len(cm.records), 1)

        child = root.getChild("child")
        self.assertEqual(child.getEffectiveLevel(), root.getEffectiveLevel())

        # The root logger could be modified by the test environment.
        # We need to pick a level that is different.
        child.setLevel(root.getEffectiveLevel() - 5)
        self.assertNotEqual(child.getEffectiveLevel(), root.getEffectiveLevel())

    def testTraceSetAt(self):
        log_name = "lsst.afw"
        root_level = logging.getLogger().getEffectiveLevel()
        trace_set_at(log_name, 2)
        trace2_log = getLogger(f"TRACE2.{log_name}")
        trace3_log = getLogger(f"TRACE3.{log_name}")
        self.assertEqual(trace2_log.getEffectiveLevel(), logging.DEBUG)
        self.assertEqual(trace3_log.getEffectiveLevel(), logging.INFO)

        # Check that child loggers are affected.
        log_name = "lsst.daf"
        child3_log = getLogger("TRACE3.lsst.daf")
        child2_log = getLogger("TRACE2.lsst.daf")
        self.assertEqual(child3_log.getEffectiveLevel(), root_level)
        self.assertEqual(child2_log.getEffectiveLevel(), root_level)
        trace_set_at("lsst", 2)
        self.assertEqual(child3_log.getEffectiveLevel(), logging.INFO)
        self.assertEqual(child2_log.getEffectiveLevel(), logging.DEBUG)

        # Also check the root logger.
        trace_set_at("", 3)
        self.assertEqual(trace3_log.getEffectiveLevel(), logging.INFO)
        self.assertEqual(getLogger("TRACE3.test").getEffectiveLevel(), logging.DEBUG)

    def test_periodic(self):
        logger = getLogger("test.periodicity")
        periodic = PeriodicLogger(logger)

        # First message will not be issued.
        periodic.log("Message")
        self.assertEqual(periodic.num_issued, 0)

        # Create a new periodic logger with no delay.
        # Every message should be issued.
        periodic = PeriodicLogger(logger, interval=0.0)
        with self.assertLogs(logger.name, level=logger.VERBOSE) as cm:
            periodic.log("Message")
            periodic.log("Message %d", 1)
        self.assertEqual(len(cm.output), 2)
        self.assertEqual(periodic.num_issued, 2)
        self.assertEqual(cm.output[0], f"VERBOSE:{logger.name}:Message")
        self.assertEqual(cm.output[1], f"VERBOSE:{logger.name}:Message 1")
        self.assertEqual(cm.records[0].filename, "test_logging.py", str(cm.records[0]))

        # Create a new periodic logger with small delay.
        # One message should be issued.
        periodic = PeriodicLogger(logger, interval=0.2, level=logger.INFO)
        with self.assertLogs(logger.name, level=logger.INFO) as cm:
            periodic.log("Message")
            time.sleep(0.5)
            issued = periodic.log("Message %d", 1)
            self.assertTrue(issued)
            issued = periodic.log("Message %d", 2)
            self.assertFalse(issued)
        self.assertEqual(periodic.num_issued, 1)
        self.assertEqual(cm.output[0], f"INFO:{logger.name}:Message 1")

        # Again with a standard python Logger.
        pylog = logging.getLogger("python.logger")
        periodic = PeriodicLogger(pylog, interval=0.0, level=logging.DEBUG)
        with self.assertLogs(pylog.name, level=logging.DEBUG) as cm:
            periodic.log("Message")
        self.assertEqual(cm.records[0].filename, "test_logging.py", str(cm.records[0]))


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import unittest

import lsst.utils
import lsst.utils.tests


class DeprecatedTestCase(lsst.utils.tests.TestCase):
    """Test depreaction."""

    def test_deprecate_pybind11(self):
        def old(x):
            """Docstring"""
            return x + 1

        # Use an unusual category
        old = lsst.utils.deprecate_pybind11(
            old, reason="For testing.", version="unknown", category=PendingDeprecationWarning
        )
        with self.assertWarnsRegex(
            PendingDeprecationWarning,
            r"Call to deprecated function \(or staticmethod\) old\. \(For testing\.\) "
            "-- Deprecated since version unknown.$",
        ):
            # Check that the function still works
            self.assertEqual(old(3), 4)
        self.assertIn("Docstring", old.__doc__)
        self.assertIn("For testing.", old.__doc__)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import unittest

import lsst.utils.tests


class TestCaseOrdering(lsst.utils.tests.TestCase):
    """Test that tests are ordered as expected."""

    def testTestOrdering(self):
        class DummyTest(lsst.utils.tests.TestCase):
            def noOp(self):
                pass

        class DummyMemoryTest(lsst.utils.tests.MemoryTestCase):
            pass

        class DummyTest2(unittest.TestCase):
            def noOp(self):
                pass

        suite = unittest.defaultTestLoader.suiteClass(
            [DummyTest2("noOp"), DummyMemoryTest("testFileDescriptorLeaks"), DummyTest("noOp")]
        )

        self.assertNotIsInstance(suite._tests[0], lsst.utils.tests.MemoryTestCase)
        self.assertIsInstance(suite._tests[-1], lsst.utils.tests.MemoryTestCase)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.
#

import unittest

import numpy as np
from lsst.utils.plotting.limits import calculate_safe_plotting_limits, make_calculate_safe_plotting_limits


class PlottingLimitsClosureTestCase(unittest.TestCase):
    """Tests for `make_calculate_safe_plotting_limits` function."""

    xs = np.linspace(0, 10, 10000)
    series1 = np.sin(xs + 3.1415 / 2) + 0.75  # min=-0.24999, max=1.74999
    series1_min = min(series1)
    series1_max = max(series1)

    series2 = np.sin(xs) + 1.2  # min=0.2, max=2.19999
    series2_min = min(series2)
    series2_max = max(series2)

    outliers = series1[:]
    outliers[1000] = 20
    outliers[2000] = -1000

    def testSingleSeries(self):
        """Test that a single series works and the outliers exclusion works."""
        # Deliberately test the bounds are the same when using the series
        # itself, and the copy with the outlier values, i.e. using
        # self.series1_min/max inside the loop despite changing the series we
        # loop over is the intent here, not a bug.
        for series in [self.series1, self.outliers]:
            ymin, ymax = make_calculate_safe_plotting_limits()(series)
            self.assertLess(ymin, self.series1_min)
            self.assertGreater(ymin, self.series1_min - 1)
            self.assertLess(ymax, self.series1_max + 1)
            self.assertGreater(ymax, self.series1_max)

    def testMultipleSeries(self):
        """Test that passing multiple several series in works wrt outliers."""
        calculate_safe_plotting_limits_accumulator = make_calculate_safe_plotting_limits()
        _, _ = calculate_safe_plotting_limits_accumulator(self.series1)
        ymin, ymax = calculate_safe_plotting_limits_accumulator(self.outliers)

        self.assertLess(ymin, self.series1_min)
        self.assertGreater(ymin, self.series1_min - 1)
        self.assertLess(ymax, self.series1_max + 1)
        self.assertGreater(ymax, self.series1_max)

    def testMultipleSeriesCommonRange(self):
        """Test that passing multiple several series in works wrt outliers."""
        calculate_safe_plotting_limits_accumulator = make_calculate_safe_plotting_limits()
        _, _ = calculate_safe_plotting_limits_accumulator(self.series1)
        ymin, ymax = calculate_safe_plotting_limits_accumulator(self.series2)
        # lower bound less than the lowest of the two
        self.assertLess(ymin, min(self.series1_min, self.series2_min))
        # lower bound less than the lowest of the two, but not by much
        self.assertGreater(ymin, min(self.series1_min, self.series2_min) - 1)
        # upper bound greater than the highest of the two
        self.assertGreater(ymax, max(self.series1_max, self.series2_max))
        # upper bound greater than the highest of the two, but not by much
        self.assertLess(ymax, max(self.series1_max, self.series2_max) + 1)

    def testSymmetric(self):
        """Test that the symmetric option works"""
        calc = make_calculate_safe_plotting_limits(symmetric_around_zero=True)
        _, _ = calc(self.series1)
        ymin, ymax = calc(self.outliers)

        self.assertEqual(ymin, -ymax)
        self.assertGreater(ymax, self.series1_max)
        self.assertLess(ymin, self.series1_min)

    def testConstantExtra(self):
        """Test that the constantExtra option works"""
        calc = make_calculate_safe_plotting_limits(constant_extra=0)
        _, _ = calc(self.series1)
        strictMin, strictMax = calc(self.outliers)

        self.assertAlmostEqual(strictMin, self.series1_min, places=4)
        self.assertAlmostEqual(strictMax, self.series1_max, places=4)

        for extra in [-2.123, -1, 0, 1, 1.5, 23]:
            calc = make_calculate_safe_plotting_limits(constant_extra=extra)
            _, _ = calc(self.series1)
            ymin, ymax = calc(self.outliers)

            self.assertAlmostEqual(ymin, self.series1_min - extra, places=4)
            self.assertAlmostEqual(ymax, self.series1_max + extra, places=4)

    def testSeriesOfSeries(self):
        """Test that we can pass a list of series to the accumulator in one."""
        calculate_safe_plotting_limits_accumulator = make_calculate_safe_plotting_limits()
        ymin, ymax = calculate_safe_plotting_limits_accumulator([self.series1, self.outliers])

        self.assertLess(ymin, self.series1_min)
        self.assertGreater(ymin, self.series1_min - 1)
        self.assertLess(ymax, self.series1_max + 1)
        self.assertGreater(ymax, self.series1_max)

    def testRaises(self):
        with self.assertRaises(TypeError):
            make_calculate_safe_plotting_limits()(1.234)


class PlottingLimitsTestCase(unittest.TestCase):
    """Tests for `calculate_safe_plotting_limits` function."""

    xs = np.linspace(0, 10, 10000)
    series1 = np.sin(xs + 3.1415 / 2) + 0.75  # min=-0.24999, max=1.74999
    series1_min = min(series1)
    series1_max = max(series1)

    series2 = np.sin(xs) + 1.2  # min=0.2, max=2.19999
    series2_min = min(series2)
    series2_max = max(series2)

    outliers = series1[:]
    outliers[1000] = 20
    outliers[2000] = -1000

    def testSingleSeries(self):
        """Test that a single series works and the outliers exclusion works."""
        # Deliberately test the bounds are the same when using the series
        # itself, and the copy with the outlier values, i.e. using
        # self.series1_min/max inside the loop despite changing the series we
        # loop over is the intent here, not a bug.
        for series in [self.series1, self.outliers]:
            ymin, ymax = calculate_safe_plotting_limits(series)
            self.assertLess(ymin, self.series1_min)
            self.assertGreater(ymin, self.series1_min - 1)
            self.assertLess(ymax, self.series1_max + 1)
            self.assertGreater(ymax, self.series1_max)

    def testMultipleSeries(self):
        """Test that passing multiple several series in works wrt outliers."""
        ymin, ymax = calculate_safe_plotting_limits([self.series1, self.outliers])
        self.assertLess(ymin, self.series1_min)
        self.assertGreater(ymin, self.series1_min - 1)
        self.assertLess(ymax, self.series1_max + 1)
        self.assertGreater(ymax, self.series1_max)

    def testMultipleSeriesCommonRange(self):
        """Test that passing multiple several series in works wrt outliers."""
        ymin, ymax = calculate_safe_plotting_limits([self.series1, self.series2])
        # lower bound less than the lowest of the two
        self.assertLess(ymin, min(self.series1_min, self.series2_min))
        # lower bound less than the lowest of the two, but not by much
        self.assertGreater(ymin, min(self.series1_min, self.series2_min) - 1)
        # upper bound greater than the highest of the two
        self.assertGreater(ymax, max(self.series1_max, self.series2_max))
        # upper bound greater than the highest of the two, but not by much
        self.assertLess(ymax, max(self.series1_max, self.series2_max) + 1)

    def testSymmetric(self):
        """Test that the symmetric option works"""
        ymin, ymax = calculate_safe_plotting_limits([self.series1, self.outliers], symmetric_around_zero=True)
        self.assertEqual(ymin, -ymax)
        self.assertGreater(ymax, self.series1_max)
        self.assertLess(ymin, self.series1_min)

    def testConstantExtra(self):
        """Test that the constantExtra option works"""
        strictMin, strictMax = calculate_safe_plotting_limits([self.series1, self.outliers], constant_extra=0)
        self.assertAlmostEqual(strictMin, self.series1_min, places=4)
        self.assertAlmostEqual(strictMax, self.series1_max, places=4)

        for extra in [-2.123, -1, 0, 1, 1.5, 23]:
            ymin, ymax = calculate_safe_plotting_limits([self.series1, self.outliers], constant_extra=extra)
            self.assertAlmostEqual(ymin, self.series1_min - extra, places=4)
            self.assertAlmostEqual(ymax, self.series1_max + extra, places=4)

    def testRaises(self):
        with self.assertRaises(TypeError):
            calculate_safe_plotting_limits(1.234)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import datetime
import logging
import os.path
import pstats
import tempfile
import time
import unittest
from dataclasses import dataclass

from astropy import units as u
from lsst.utils.timer import logInfo, logPairs, profile, time_this, timeMethod

log = logging.getLogger("test_timer")

THIS_FILE = os.path.basename(__file__)

# Only use this in a single test but needs to be associated
# with the function up front.
test_metadata = {}


@dataclass
class Example1:
    """Test class with log and metadata properties, similar to ``Task``."""

    log: logging.Logger
    metadata: dict

    @timeMethod
    def sleeper(self, duration: float) -> None:
        """Sleep for some time."""
        time.sleep(duration)


@timeMethod
def decorated_sleeper_nothing(self, duration: float) -> None:
    """Test function that sleeps."""
    time.sleep(duration)


@timeMethod(logger=log)
def decorated_sleeper_logger(self, duration: float) -> None:
    """Test function that sleeps and logs."""
    time.sleep(duration)


@timeMethod(logger=log, logLevel=logging.INFO)
def decorated_sleeper_logger_level(self, duration: float) -> None:
    """Test function that logs at INFO."""
    time.sleep(duration)


@timeMethod(metadata=test_metadata)
def decorated_sleeper_metadata(self, duration: float) -> None:
    """Test function that uses external metadata."""
    time.sleep(duration)


class TestTimeMethod(unittest.TestCase):
    """Test the time method decorator."""

    def testLogPairs(self):
        # Test the non-obj case.
        logger = logging.getLogger("test")
        pairs = (("name1", 0), ("name2", 1))
        metadata = {}
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logLevel=logging.INFO, logger=logger, metadata=metadata)
        self.assertEqual(len(cm.output), 1, cm.output)
        self.assertTrue(cm.output[0].endswith("name1=0; name2=1"), cm.output)
        self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")
        self.assertEqual(metadata, {"name1": [0], "name2": [1]})

        # Call it again with an explicit stack level.
        # Force it to come from lsst.utils.
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logLevel=logging.INFO, logger=logger, metadata=metadata, stacklevel=0)
        self.assertEqual(cm.records[0].filename, "timer.py")

        # Check that the log message is filtered by default.
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logger=logger, metadata=metadata)
            logger.info("Message")
        self.assertEqual(len(cm.records), 1)

    def testLogInfo(self):
        metadata = {}
        logger = logging.getLogger("testLogInfo")
        with self.assertLogs(level=logging.INFO) as cm:
            logInfo(None, prefix="Prefix", metadata=metadata, logger=logger, logLevel=logging.INFO)
        self.assertEqual(cm.records[0].filename, THIS_FILE)
        self.assertIn("PrefixUtc", metadata)
        self.assertIn("PrefixMaxResidentSetSize", metadata)
        self.assertEqual(metadata["__version__"], 1)

        # Again with no log output.
        logInfo(None, prefix="Prefix", metadata=metadata)
        self.assertEqual(len(metadata["PrefixUtc"]), 2)

        # With an explicit stacklevel.
        with self.assertLogs(level=logging.INFO) as cm:
            logInfo(
                None, prefix="Prefix", metadata=metadata, logger=logger, logLevel=logging.INFO, stacklevel=0
            )
        self.assertEqual(cm.records[0].filename, "timer.py")

    def assertTimer(self, duration, task):
        # Call it twice to test the "add" functionality.
        task.sleeper(duration)
        task.sleeper(duration)
        counter = 2

        has_logger = getattr(task, "log", None) is not None and task.log is not None
        has_metadata = getattr(task, "metadata", None) is not None and task.metadata is not None

        if has_logger:
            counter += 1
            with self.assertLogs("timer.task", level=logging.DEBUG) as cm:
                task.sleeper(duration)
            self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")

        if has_metadata:
            self.assertEqual(len(task.metadata["sleeperStartUserTime"]), counter)

            start = datetime.datetime.fromisoformat(task.metadata["sleeperStartUtc"][1])
            end = datetime.datetime.fromisoformat(task.metadata["sleeperEndUtc"][1])
            delta = end - start
            delta_sec = delta.seconds + (delta.microseconds / 1e6)
            self.assertGreaterEqual(delta_sec, duration)

    def testTaskLike(self):
        """Test timer on something that looks like a Task."""
        # Call with different parameters.
        parameters = (
            (logging.getLogger("task"), {}),
            (logging.getLogger("task"), None),
            (None, {}),
            (None, None),
        )

        duration = 0.1
        for log, metadata in parameters:
            with self.subTest(log=log, metadata=metadata):
                task = Example1(log=log, metadata=metadata)
                self.assertTimer(duration, task)

    def testDecorated(self):
        """Test timeMethod on non-Task like instances."""
        duration = 0.1

        # The "self" object shouldn't be usable but this should do nothing
        # and not crash.
        decorated_sleeper_nothing(self, duration)

        # Use a function decorated for logging.
        with self.assertLogs("timer.test_timer", level=logging.DEBUG) as cm:
            decorated_sleeper_logger(self, duration)
        self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")

        # And adjust the log level
        with self.assertLogs("timer.test_timer", level=logging.INFO):
            decorated_sleeper_logger_level(self, duration)

        # Use a function decorated for metadata.
        self.assertEqual(len(test_metadata), 0)
        with self.assertLogs("timer.test_timer", level=logging.DEBUG) as cm:
            # Check that we only get a single log message and nothing from
            # timeMethod itself.
            decorated_sleeper_metadata(self, duration)
            logging.getLogger("timer.test_timer").debug("sentinel")
        self.assertEqual(len(cm.output), 1)
        self.assertIn("decorated_sleeper_metadataStartUserTime", test_metadata)


class TimerTestCase(unittest.TestCase):
    """Test the timer functionality."""

    def testTimer(self):
        with self.assertLogs(level="DEBUG") as cm:
            with time_this():
                pass
        self.assertEqual(cm.records[0].name, "timer")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertEqual(cm.records[0].filename, THIS_FILE)

        with self.assertLogs(level="DEBUG") as cm:
            with time_this(prefix=None):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertNotIn(": Took", cm.output[0])
        self.assertNotIn("; ", cm.output[0])
        self.assertEqual(cm.records[0].filename, THIS_FILE)

        # Report memory usage.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("byte", cm.output[0])

        # Request memory usage but with log level that will not issue it.
        with self.assertLogs(level="INFO") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True):
                pass
            # Ensure that a log message is issued.
            _log = logging.getLogger()
            _log.info("info")
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "INFO")
        all = "\n".join(cm.output)
        self.assertNotIn("Took", all)
        self.assertNotIn("memory", all)

        # Report memory usage including child processes.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_child=True):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("byte", cm.output[0])

        # Report memory usage, use non-default, but a valid memory unit.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_unit=u.kilobyte):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("kbyte", cm.output[0])

        # Report memory usage, use an invalid memory unit.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_unit=u.gram):
                pass
        self.assertEqual(cm.records[0].name, "lsst.utils.timer")
        self.assertEqual(cm.records[0].levelname, "WARNING")
        self.assertIn("Invalid", cm.output[0])
        self.assertIn("byte", cm.output[0])
        self.assertEqual(cm.records[1].name, "root")
        self.assertEqual(cm.records[1].levelname, "DEBUG")
        self.assertIn("Took", cm.output[1])
        self.assertIn("memory", cm.output[1])
        self.assertIn("delta", cm.output[1])
        self.assertIn("peak delta", cm.output[1])
        self.assertIn("byte", cm.output[1])

        # Change logging level
        with self.assertLogs(level="INFO") as cm:
            with time_this(level=logging.INFO, prefix=None):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertIn("Took", cm.output[0])
        self.assertIn("seconds", cm.output[0])

        # Use a new logger with a message.
        msg = "Test message %d"
        test_num = 42
        logname = "test"
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(log=logging.getLogger(logname), msg=msg, args=(42,), prefix=None):
                pass
        self.assertEqual(cm.records[0].name, logname)
        self.assertIn("Took", cm.output[0])
        self.assertIn(msg % test_num, cm.output[0])

        # Prefix the logger.
        prefix = "prefix"
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(prefix=prefix):
                pass
        self.assertEqual(cm.records[0].name, prefix)
        self.assertIn("Took", cm.output[0])

        # Prefix explicit logger.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(log=logging.getLogger(logname), prefix=prefix):
                pass
        self.assertEqual(cm.records[0].name, f"{prefix}.{logname}")

        # Trigger a problem.
        with self.assertLogs(level="ERROR") as cm:
            with self.assertRaises(RuntimeError):
                with time_this(log=logging.getLogger(logname), prefix=prefix):
                    raise RuntimeError("A problem")
        self.assertEqual(cm.records[0].name, f"{prefix}.{logname}")
        self.assertEqual(cm.records[0].levelname, "ERROR")


class ProfileTestCase(unittest.TestCase):
    """Test profiling decorator."""

    def test_profile(self):
        logger = logging.getLogger("profile")

        with profile(None) as prof:
            pass
        self.assertIsNone(prof)

        with tempfile.NamedTemporaryFile() as tmp:
            with self.assertLogs("profile", level=logging.INFO) as cm:
                with profile(tmp.name, logger) as prof:
                    pass
            self.assertEqual(len(cm.output), 2)
            self.assertIsNotNone(prof)
            self.assertTrue(os.path.exists(tmp.name))
            self.assertIsInstance(pstats.Stats(tmp.name), pstats.Stats)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import os
import sys
import unittest

import lsst.utils.tests
from lsst.utils import getPackageDir


@unittest.skipIf("UTILS_DIR" not in os.environ, "EUPS has not set up this package.")
class GetPackageDirTestCase(unittest.TestCase):
    """Test that EUPS environment variables can be read."""

    def testBasics(self):
        utilsPath = getPackageDir("utils")
        self.assertTrue(os.path.isfile(os.path.join(utilsPath, "tests", "test_getPackageDir.py")))

        # Confirm that we have a correct Python exception and pex exception
        with self.assertRaises(LookupError):
            getPackageDir("nameOfNonexistendPackage2234q?#!")

    def testUnicodeBasics(self):
        utilsPath = getPackageDir("utils")
        self.assertTrue(os.path.isfile(os.path.join(utilsPath, "tests", "test_getPackageDir.py")))


class TestMemory(lsst.utils.tests.MemoryTestCase):
    """Test for file descriptor leaks."""


def setup_module(module):
    """Initialize the pytest environment."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    setup_module(sys.modules[__name__])
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import sys
import unittest
from collections import Counter

# Classes and functions to use in tests.
import lsst.utils
from lsst.utils import doImport
from lsst.utils._packaging import getPackageDir
from lsst.utils.introspection import (
    find_outside_stacklevel,
    get_caller_name,
    get_class_of,
    get_full_type_name,
    get_instance_of,
)


class GetCallerNameTestCase(unittest.TestCase):
    """Test get_caller_name

    Warning: due to the different ways this can be run
    (e.g. directly or py.test), the module name can be one of two different
    things.
    """

    def test_free_function(self):
        def test_func():
            return get_caller_name(1)

        result = test_func()
        self.assertEqual(result, f"{__name__}.test_func")

    def test_instance_method(self):
        class TestClass:
            def run(self):
                return get_caller_name(1)

        tc = TestClass()
        result = tc.run()
        self.assertEqual(result, f"{__name__}.TestClass.run")

    def test_class_method(self):
        class TestClass:
            @classmethod
            def run(cls):
                return get_caller_name(1)

        tc = TestClass()
        result = tc.run()
        self.assertEqual(result, f"{__name__}.TestClass.run")

    def test_skip(self):
        def test_func(stacklevel):
            return get_caller_name(stacklevel)

        result = test_func(2)
        self.assertEqual(result, f"{__name__}.GetCallerNameTestCase.test_skip")

        result = test_func(2000000)  # use a large number to avoid details of how the test is run
        self.assertEqual(result, "")


class TestInstropection(unittest.TestCase):
    """Tests for lsst.utils.introspection."""

    def testTypeNames(self):
        # Check types and also an object
        tests = [
            (getPackageDir, "lsst.utils.getPackageDir"),  # underscore filtered out
            (int, "int"),
            (0, "int"),
            ("", "str"),
            (doImport, "lsst.utils.doImport.doImport"),  # no underscore
            (Counter, "collections.Counter"),
            (Counter(), "collections.Counter"),
            (lsst.utils, "lsst.utils"),
        ]

        for item, typeName in tests:
            self.assertEqual(get_full_type_name(item), typeName)

    def testUnderscores(self):
        # Underscores are filtered out unless they can't be, either
        # because __init__.py did not import it or there is a clash with
        # the non-underscore version.
        for test_name in (
            "import_test.two._four.simple.Simple",
            "import_test.two._four.clash.Simple",
            "import_test.two.clash.Simple",
        ):
            test_cls = get_class_of(test_name)
            self.assertTrue(test_cls.true())
            full = get_full_type_name(test_cls)
            self.assertEqual(full, test_name)

    def testGetClassOf(self):
        tests = [(doImport, "lsst.utils.doImport"), (Counter, "collections.Counter")]

        for test in tests:
            ref_type = test[0]
            for t in test:
                c = get_class_of(t)
                self.assertIs(c, ref_type)

    def testGetInstanceOf(self):
        c = get_instance_of("collections.Counter", "abcdeab")
        self.assertIsInstance(c, Counter)
        self.assertEqual(c["a"], 2)
        with self.assertRaises(TypeError) as cm:
            get_instance_of(lsst.utils)
        self.assertIn("lsst.utils", str(cm.exception))

    def test_stacklevel(self):
        level = find_outside_stacklevel("lsst.utils")
        self.assertEqual(level, 1)

        info = {}
        level = find_outside_stacklevel("lsst.utils", stack_info=info)
        self.assertIn("test_introspection.py", info["filename"])

        c = doImport("import_test.two.three.success.Container")
        with self.assertWarns(Warning) as cm:
            level = c.level()
        self.assertTrue(cm.filename.endswith("test_introspection.py"))
        self.assertEqual(level, 2)
        with self.assertWarns(Warning) as cm:
            level = c.indirect_level()
        self.assertTrue(cm.filename.endswith("test_introspection.py"))
        self.assertEqual(level, 3)

        # Test with additional options.
        with self.assertWarns(Warning) as cm:
            level = c.indirect_level(allow_methods={"indirect_level"})
        self.assertEqual(level, 2)
        self.assertTrue(cm.filename.endswith("success.py"))

        # Adjust test on python 3.10.
        allow_methods = {"import_test.two.three.success.Container.level"}
        stacklevel = 1
        if sys.version_info < (3, 11, 0):
            # python 3.10 does not support "." syntax and will filter it out.
            allow_methods.add("indirect_level")
            stacklevel = 2
        with self.assertWarns(FutureWarning) as cm:
            level = c.indirect_level(allow_methods=allow_methods)
        self.assertEqual(level, stacklevel)
        self.assertTrue(cm.filename.endswith("success.py"))


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import unittest

from lsst.utils.tests import ImportTestCase


class TestImports(ImportTestCase):
    """Test we can run the import tests.

    All the code in utils is being tested and imported already but this
    will confirm the test infrastructure works.
    """

    PACKAGES = ("lsst.utils",)

    def test_counter(self):
        """Test that the expected number of packages were registered."""
        self.assertEqual(self._n_registered, 1)


if __name__ == "__main__":
    unittest.main()
